/*
  Command processing robot using Arduino Pro Mini with RMx range finder
  US droid mini robot sample sketch

  Created by Steve Nishiarai 18/Jan/2019
  Copyright (c) 2019 Parawind Lab.  All right reserved.
*/

#define VERSION "MB_190128"

#include <EEPROM.h>
//for using Serial
#include <stdio.h>
//for using sprintf
#include <stdlib.h>
// I2C Master library
#include <Wire.h>
// MiniBot motor control library
#include "MBmotor.h"

/****************************************************************
 * Serial Interface commands
 *
 * Command is "$" followd by a single Ascii letter character and followed by six or two characters which represents hexa code.
 * Following are the command list. Other than command letter, "c" represents alphabet one character, "h" represents one character of hexa code and "d" represents one character of decimal code.
 * Command and arguments are echoed back followed by <CR><LF> after the command is accepted.
 *
 *List of command letters: A,B,E,f,F,H,I,l,L,M,Q,r,R,S,T,V,W,X,Y

 *Choreograph start/stop
 *------------------------------------------------------------------------------
 *  12
 *  ||
 * $A;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge

 *Read EEPROM data for 16 bytes unit
 *------------------------------------------------------------------------------
 *  12  34
 *  ||  ||
 * $Ehh;hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh,hh<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (EEPROM upper address 0x00 - 0x3F)
 *3. Acknowledge
 *4. EEPROM data (16 bytes)
 *

 *Go forward by range settings in 100 millisecond unit
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $fhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Moving time. hh * 100 millisecond)
 *3. Acknowledge

 *Go forward by range settings in centimeter
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $Fhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Running length in centimeter). if value is 0 then run continuously.
 *3. Acknowledge

 *Send command to RMx sensor
 *------------------------------------------------------------------------------
 *  12 34  5   6
 *  || ||  |   |
 * $Ihhch-h<LF>;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. I2C address
 *3. Command to RMx
 *4. Argument to RMx
 *5. Termination of argument (LF code)
 *6. Acknowledge
 *

 *Turn left by angle settings in 100 millisecond unit
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $lhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Moving time. hh * 100 millisecond)
 *3. Acknowledge

 *Turn left by angle settings in degree
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $Lhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Turning angle in degree). if value is 0 then run continuously.
 *3. Acknowledge

 *Read from RMx sensor
 *------------------------------------------------------------------------------
 *  12 345    6
 *  || |||    |
 * $Qhh;cdddd dd<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. I2C address
 *3. Acknowledge
 *4. Indication code from RMx
 *5. Measurement result of distance from RMx
 *6. Measurement result of intensity from RMx
 *

 *Turn right by angle settings in 100 millisecond unit
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $rhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Moving time. hh * 100 millisecond)
 *3. Acknowledge

 *Turn right by angle settings in degree
 *------------------------------------------------------------------------------
 *  12 3
 *  || |
 * $Rhh;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Argument (Turning angle in degree). if value is 0 then run continuously.
 *3. Acknowledge

 *on/off of the continuous read from RMx sensor
 *------------------------------------------------------------------------------
 *  12        34    5
 *  ||        ||    |
 * $S;<CR><LF>cdddd dd<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge
 *3. Indication code from RMx
 *4. Measurement result distance from RMx
 *5. Measurement result of intensity from RMx
 *
 * 3,4,5 and <CR><LF> repeats until receiving this command again
 * I2C address is assigned by last time 'Q' command.

 *Show version number
 *------------------------------------------------------------------------------
 *  12        3
 *  ||        |
 * $V;<CR><LF>dddd<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge
 *3. Version number

 *Stop motor
 *------------------------------------------------------------------------------
 *  12
 *  ||
 * $T;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge

 *Write EEPROM data
 *------------------------------------------------------------------------------
 *  12  34  5 67
 *  ||  ||  | ||
 * $Whhh:hh,hh.;<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. EEPROM address up to 3FF
 *3. Divider between address and data array
 *4. Writing data
 *5. Second writing data
 *6. Terminator of the writing data
 *7. Acknowledge
 *
 * Multiple data can be send divided by "," and should be terminated by ".".
 * The number of transferring characters must be less or equal to 64, since the size of serial buffer is 64 bytes.

 *Read battery voltage
 *------------------------------------------------------------------------------
 *  12        3
 *  ||        |
 * $X;<CR><LF>d.dV<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge
 *3. Battery voltage

 *Show lowest battery voltage when motor was driven last time
 *------------------------------------------------------------------------------
 *  12        3
 *  ||        |
 * $Y;<CR><LF>d.dV<CR><LF>
 *------------------------------------------------------------------------------
 *
 *1. Command echo back
 *2. Acknowledge
 *3. Battery voltage

*/
//***************************************************************

/****************************************************************
 * Remote control commands for the purpose of fit to the Arduino Bluetooth Controller by Ioannis Tzanellis
*/

#define RC_COM_FORWARD  'f'
#define RC_COM_STEERL   'l'
#define RC_COM_STEERR   'r'
#define RC_COM_STOP     's'
#define RC_COM_TURNL    't'
#define RC_COM_TURNR    'u'
#define RC_COM_ACTION   'a'

#define RC_STEER_FACTOR  100
//Time of the stop after the last command in millisecond
#define RC_TIMEOUT  200
//Previous command buffer for detecting the command change
char PrevRCcommand = ' ' ;
//Timestamp buffer for the previous command received time
long TimeRCcommand = 0 ;

//***************************************************************

/****************************************************************
 * EEPROM Map
 *
 *------------------------------------------------------------------------------
 * Address: 0
 * Initialization flag
 *
 *  If this byte is 0xFF then initialize all configuration data to default at the start up.
 *
*/
#define EEADR_FLAG 0
#define INIT_FLAG 0x00

/*------------------------------------------------------------------------------
 * Address: 1
 * Trim data of the right motor when moving forward
 *
 * the value is signed 8 bit. (-128 : 127)
*/
#define EEADR_MR_FTRIM 1
#define INIT_MR_FTRIM 0

/*------------------------------------------------------------------------------
 * Address: 2
 * Trim data of left motor when moving forward
 *
 * the value is signed 8 bit. (-128 : 127)
*/
#define EEADR_ML_FTRIM 2
#define INIT_ML_FTRIM 0

/*------------------------------------------------------------------------------
 * Address: 3
 * Trim data of right motor at left turn
 *
 * the value is signed 8 bit. (-128 : 127)
*/
#define EEADR_MR_TTRIM 3
#define INIT_MR_TTRIM 0

/*------------------------------------------------------------------------------
 * Address: 4
 *Trim data of left motor at right turn
 *
 * the value is signed 8 bit. (-128 : 127)
*/
#define EEADR_ML_TTRIM 4
#define INIT_ML_TTRIM 0

/*------------------------------------------------------------------------------
 * Address: 5
 * Minimum speed
*/
#define EEADR_MINSPEED 5
#define INIT_MINSPEED 0x03

/*------------------------------------------------------------------------------
 * Address: 6
 * Nominal speed
*/
#define EEADR_NOMSPEED 6
#define INIT_NOMSPEED 0x10

/*------------------------------------------------------------------------------
 * Address: 7
 * Acceleration time in 10 milli second unit
 *
*/
#define EEADR_ACCELTIME 7
#define INIT_ACCELTIME 50

/*------------------------------------------------------------------------------
 * Address: 8
 * Obstacle detection range
 *
 *  Unit is 1cm, terefore the maximum range is 255cm
*/
#define EEADR_OBSTACLE_RANGE 8
#define INIT_OBSTACLE_RANGE 20
unsigned int ObstacleRange ;

/*------------------------------------------------------------------------------
 * Address: 9
 * Battery Low detection voltage (vlotage * 256 / 5)
*/
#define EEADR_BATTLOW 9
#define INIT_BATTLOW 102 //2V

/*------------------------------------------------------------------------------
 * Address: 10
 * Motor feedback gain
*/
#define EEADR_MTFBGAIN 10
#define INIT_MTFBGAIN 40

/*------------------------------------------------------------------------------
 * Address: 11 - 63
 * Reserved
*/

/*------------------------------------------------------------------------------
 * Address: 64 - 1023
 * Choreograph commands
 *
 * Choreograph command is consisted of 1 byte of command code followed by 0 or 1 byte of parameters.
 * Number of parameter is dependent on each command.
 * Choreograph sequence is start from address 64 and is stop by the terminator command of 0xFF.
 *
 * Choreograph commnad list
 *
 * +---------Command code
 * |    +----Number of parameters
 * |    | +--Brief
 * |    | |
 * 0x00 1 Change obstacle detection range to the parameter value in cm.
 * 0x01 1 Change Choreograph stepping time in 100mS unit
 * 0x10 1 Jump to the address of the parameter.
 * 0x11 1 Jump to the address of the parameter + 256.
 * 0x12 1 Jump to the address of the parameter + 512.
 * 0x13 1 Jump to the address of the parameter + 768.
 * 0x14 1 Jump to the address of the parameter, if the obstacle is found.
 * 0x15 1 Jump to the address of the parameter + 256, if the obstacle is found.
 * 0x16 1 Jump to the address of the parameter + 512, if the obstacle is found.
 * 0x17 1 Jump to the address of the parameter + 768, if the obstacle is found.
 * 0x20 1 Pause for time of parameter multiplied by 100mS.
 * 0x21 1 Proceed forward by the cm unit parameter.(Stop and quit when obstacle is detected.)
 * 0x22 1 Turn right by the degree unit parameter.
 * 0x23 1 Turn left by the degree unit parameter.
 * 0xF0 0 Restart command of the choreograph sequence.(Jump to address 0x40)
 * 0xF1 0 Terminator of the choreograph sequence.
 * 0xFE 0 Read distance to the obstacle in front.(Distance is read in the buffer.)
 * 0xFF 0 No operation.
*/
#define CHORESEQADR 64
unsigned int ChoreSeqAdr = CHORESEQADR ;
unsigned int DistanceBuf ;
unsigned int ChoreSeqCount ;
//Choreograph stepping time (100mS unit)
byte ChoreStepTime = 0 ;
//***************************************************************

//////Flags//////
boolean FlagChoreograph = false ;
boolean FlagSens = false ;

//////Continuous running state//////
#define CONT_OFF 0
#define CONT_RIGHT 1
#define CONT_LEFT 2
#define CONT_FORWARD 3
byte StateContRun = CONT_OFF ;

//////I2C interface//////
#define I2CADR0 0x09
//reading buffer for RMx
byte I2Cadr ;
char Id ;
byte SensLevel ;

////// Register for sensor test //////
unsigned long lastTime = 0;

////// assign motor control pins //////
#define MCL 9
#define MCR 10
//A0 is MotorL BEMF sensing pin
#define MSL 0
//A1 is MotorR BEMF sensing pin
#define MSR 1
//A2 is Motor reference sensing pin
#define MREF 2

MBmotor MyMBmotor(MCL, MCR, MSL, MSR, MREF);

//***************************************************************
/***********************************************
 * Receive three Hex characters from serial and transform to unsigned int code
 */
unsigned int Receive3CharToInt()
{
  char arg = ReceiveSerial();
  //echo back
  Serial.print(arg);
  unsigned int i = Hascii2bin(arg) ;
  i <<= 4 ;
  arg = ReceiveSerial();
  //echo back
  Serial.print(arg);
  i |= Hascii2bin(arg) ;
  i <<= 4 ;
  arg = ReceiveSerial();
  //echo back
  Serial.print(arg);
  i |= Hascii2bin(arg) ;

  return i ;
}

/***********************************************
 * Receive two Hex characters from serial and transform to byte code
 */
byte Receive2CharToByte()
{
  char arg = ReceiveSerial();
  //echo back
  Serial.print(arg);
  byte b = Hascii2bin(arg) ;
  b <<= 4 ;
  arg = ReceiveSerial();
  //echo back
  Serial.print(arg);
  b |= Hascii2bin(arg) ;

  return b ;
}

/***********************************************
 * Wait for one character but ignore space codes
 */
char ReceiveSerial()
{
  char c ;
  while (1)
  {
    while (! Serial.available()){}
    c = Serial.read();
    if ( c != ' ' && c != '  ' ) { break ;}
  }
  return c ;
}
/***********************************************
 * Read sensor value from range finder
 */
unsigned int ReadSensor(unsigned int I2Cadress)
{
  //put maximum value for the default
  unsigned int reading = 0xFFFF ;
  // request 4 bytes to sensor1
  Wire.requestFrom(I2Cadress, 4);
  //Wait I2C response for 3 milli seconds
  unsigned long EndTime = millis() + 3 ;
  while ( millis() < EndTime )
  {
    // receive reading from sensor
    if (4 <= Wire.available())   // if all four bytes were received
    {
      // receive reading from sensor
      Id = Wire.read();  //receive indication code
      reading = Wire.read();  // receive high byte (overwrites previous reading)
      reading <<= 8 ;    // shift high byte to be high 8 bits
      reading |= Wire.read(); // receive low byte as lower 8 bits
      SensLevel = Wire.read();  // receive sensing level
      break ;
    }
  }
  return reading ;
}

/***********************************************
 * Functions for serial communication
 */
byte Hascii2bin(byte input)
{
  //if the bit6 is 1 then 
  if ( input >= 0x40 )
  {
    //it must be 0x0A - 0x0F
    input += 9 ;
  }
  input &= 0x0F ;
  return input ;
}

/***********************************************
 * Show unsigned int value to the serial
 */
void ShowIntValue(unsigned int IntValue)
{
  char string[5] ;
  sprintf ( string, "%04X", IntValue ) ;
  Serial.print(string);
}

/***********************************************
 * Motion control functions
 ***********************************************/

/***********************************************
 * Calculate motor driving time from proceed distance
 *
 * From the actual measurement results, the functions of the driving time and the turning angle are as follows.
 * t= L*9.27    ; t: driving time, L: proceed distance in m
 * Since the time setting value is given in units of 1ms and distance setting value is given in units of cm, the relational expression between the time setting value and the proceed distance is as follows.
 * x=92.7c        ; x: time setting value in milli second, c:distance setting value in cm
 *
 * Parameter settings in actual measurement is as follows.
 * Nominal speed: 0x10, Acceleration time: 50
 */
unsigned int GetProcTimeFromCm( unsigned int c )
{
  float x = 92.7*(float)c ;
  return (int)x ;
}

/***********************************************
 * Calculate motor driving time from turning degree
 *
 * From the actual measurement results, the functions of the driving time and the turning angle are as follows.
 * t= 0.00735d      ; t: driving time, d: turning angle in degree
 * Since the time setting value is given in units of 1ms, the relational expression between the time setting value and the turning angle is as follows.
 * x= 7.35d          ; x: time setting value in milli second
 *
 * Parameter settings in actual measurement is as follows.
 * Nominal speed: 0x30, Acceleration time: 50
 */
unsigned int GetTurnTimeFromDeg( unsigned int d )
{
  float x = 7.35*(float)d ;
  return (int)x ;
}

/***********************************************
 * Wait for assigned time unless any obstacle found
 */
void WaitForObstacle( unsigned int RunningTime )
{
  unsigned long EndTime = millis() + (unsigned long)RunningTime ;
  while ( millis() < EndTime )
  {
    //Measure the range of obstacle in front
    unsigned int Reading = ReadSensor ( I2CADR0 ) ;
    //if there is obstacle
    if ( Reading <= ObstacleRange )
    {
      Serial.println(" Obstacle found");
      //Leave waiting loop if obstacle is found
      break ;
    }
  }
}

void GoForward( unsigned int RunningTime )
{
  //Go when running time is not 0 and battery is OK
  if ( RunningTime > 0 && CheckBattery() == true )
  {
    //if still running then wait for stop
    MyMBmotor.WaitForStop() ;
    //Proceed forward with 0 steering facotr
    MyMBmotor.ProceedForward(0,0) ;
    WaitForObstacle(RunningTime) ;
    MyMBmotor.SlowStop() ;
  }
}

void TurnRight( unsigned int RunningTime )
{
  //Go when running time is not 0 and battery is OK
  if ( RunningTime > 0 && CheckBattery() == true )
  {
    //if still running then wait for stop
    MyMBmotor.WaitForStop() ;
    MyMBmotor.RotateRight() ;
    delay(RunningTime) ;
    MyMBmotor.SlowStop() ;
  }
}

void TurnLeft( unsigned int RunningTime )
{
  //Go when running time is not 0 and battery is OK
  if ( RunningTime > 0 && CheckBattery() == true )
  {
    //if still running then wait for stop
    MyMBmotor.WaitForStop() ;
    MyMBmotor.RotateLeft() ;
    delay(RunningTime) ;
    MyMBmotor.SlowStop() ;
  }
}

/***********************************************
 * Check battery level and return true if level is OK else show voltage and return false
 */
boolean CheckBattery()
{
  //multiple by 4 since AD reading is 10 bits and low voltage setting is 8 bits
  unsigned int LowVoltage = (unsigned int)EEPROM.read(EEADR_BATTLOW) * 4 ;
  unsigned int BattVolt = MyMBmotor.ReadLowBatt() ;
  if ( BattVolt <= LowVoltage )
  {
    float Voltage = (float)BattVolt * 5 / 1024 ;
    Serial.println();
    Serial.print(Voltage);
    Serial.println("V");

    return false ;
  }
  return true ;
}

/***********************************************
 * Initialize MBmotor with parameters in EEPROM
 */
void InitMBmotor()
{
  byte MR_FTrim = EEPROM.read(EEADR_MR_FTRIM) ;
  byte ML_FTrim = EEPROM.read(EEADR_ML_FTRIM) ;
  byte MR_TTrim = EEPROM.read(EEADR_MR_TTRIM) ;
  byte ML_TTrim = EEPROM.read(EEADR_ML_TTRIM) ;
  byte MinSpeed = EEPROM.read(EEADR_MINSPEED) ;
  byte NomSpeed = EEPROM.read(EEADR_NOMSPEED) ;
  byte AccTime = EEPROM.read(EEADR_ACCELTIME) ;
  byte Gain = EEPROM.read(EEADR_MTFBGAIN) ;
  MyMBmotor.begin(MR_FTrim, ML_FTrim, MR_TTrim, ML_TTrim, MinSpeed, NomSpeed, AccTime, Gain) ;
}

/***********************************************
 * 
 */
void setup()
{
  // join i2c bus (address optional for master)
  Wire.begin();
  // start serial communication at 115200bps
  Serial.begin(115200);

  //if the first byte of EEPROM is 0xFF then initialize to default
  if ( EEPROM.read(EEADR_FLAG) == 0xFF )
  {
    EEPROM.write(EEADR_FLAG, INIT_FLAG);
    EEPROM.write(EEADR_MR_FTRIM, INIT_MR_FTRIM);
    EEPROM.write(EEADR_ML_FTRIM, INIT_ML_FTRIM);
    EEPROM.write(EEADR_MR_TTRIM, INIT_MR_TTRIM);
    EEPROM.write(EEADR_ML_TTRIM, INIT_ML_TTRIM);
    EEPROM.write(EEADR_MINSPEED, INIT_MINSPEED);
    EEPROM.write(EEADR_NOMSPEED, INIT_NOMSPEED);
    EEPROM.write(EEADR_ACCELTIME, INIT_ACCELTIME);
    EEPROM.write(EEADR_OBSTACLE_RANGE, INIT_OBSTACLE_RANGE);
    EEPROM.write(EEADR_BATTLOW, INIT_BATTLOW);
    EEPROM.write(EEADR_MTFBGAIN, INIT_MTFBGAIN);
  }

  ObstacleRange = (unsigned int)EEPROM.read(EEADR_OBSTACLE_RANGE) * 10 ;
  InitMBmotor() ;

  //Initialize I2C address
  I2Cadr = I2CADR0 ;

  //Set choreograph flag first
  FlagChoreograph = true ;
  //give time to communication device gets ready
  delay (100) ;

  //prepare receiving buffer
  char c[3] ;
  //Test RN42 communication by trying get into command mode
  Serial.print("$$$");
  //wait a moment for receive response
  delay (100) ;
  while (Serial.available())
  {
    c[0] = c[1] ;
    c[1] = c[2] ;
    c[2] = Serial.read();
    //if response is CMD
    if ( c[0] == 'C' && c[1] == 'M' && c[2] == 'D' )
    {
      //Reset choreograph flag
      FlagChoreograph = false ;
      //Leave command mode
      Serial.println("---");
      //wait a moment for receive response
      delay (100) ;
      while (Serial.available())
      {
        //dummy read
        c[2] = Serial.read();
      }
      goto CHKEND ;
    }
  }

  //Change baudrate
  Serial.begin(9600);
  //Test HC-06 communication by sending AT command
  Serial.println(); //necessary for clearing command buffer of the HC-06.
  Serial.println("AT");
  //wait a moment for receive response
  delay (100) ;
  while (Serial.available())
  {
    c[0] = c[1] ;
    c[1] = Serial.read();
    //if response is OK
    if ( c[0] == 'O' && c[1] == 'K' )
    {
      //Reset choreograph flag
      FlagChoreograph = false ;
    }
  }
  CHKEND:;

  //ignore any other incoming messages for half second since some app may sends opening messages
  delay (500) ;
  while (Serial.available())
  {
    //dummy read
    c[0] = Serial.read();
  }
}


void loop()
{
  // if some character comes from serial
  while (Serial.available())
  {
    char c = Serial.read();
    //echo back
    Serial.print(c);
    //Process command only when the indication of '$' comes
    if ( c == '$' )
    {
      c = ReceiveSerial();
      //echo back
      Serial.print(c);
      if ( c == 'W' )
      //write to EEPROM
      {
        unsigned int Adr = Receive3CharToInt();
        while (1)
        {
          char arg[2] ;
          c = ReceiveSerial();
          //echo back
          Serial.print(c);
          if ( c != ',' && c != '.' )
          {
            arg[1] = arg[0] ; arg[0] = c ;
          }
          else
          {
            byte b = Hascii2bin(arg[1]) ;
            b <<= 4 ;
            byte Data = b | Hascii2bin(arg[0]) ;
            EEPROM.write(Adr, Data);
            Adr++ ;
          }
          if ( c == '.' )
          {
            Serial.println(";");
            break;
          }
        }
        //renew obstacle detection range parameter, since it may changed
        ObstacleRange = (unsigned int)EEPROM.read(EEADR_OBSTACLE_RANGE) * 10 ;
        //restart MBmotor since motor control parameters may changed
        InitMBmotor() ;
      }
      else if ( c == 'f' )
      {
        byte b = Receive2CharToByte();
        unsigned int RunningTime = (unsigned int)b * 10 ; //multiply to 10mS resolution
        GoForward( RunningTime );
        Serial.println(";");
      }
      else if ( c == 'F' )
      {
        byte b = Receive2CharToByte();
        if ( b > 0 )
        {
          unsigned int RunningTime = GetProcTimeFromCm( (unsigned int)b ) ;
          GoForward( RunningTime );
        }
        else
        {
          //if RunningTime == 0 then start continuous run
          StateContRun = CONT_FORWARD ;
          //When continuous running starts, the choreography must be stopped
          FlagChoreograph = false ;
          MyMBmotor.ProceedForward(0,0) ;
        }
        Serial.println(";");
      }
      else if ( c == 'r' )
      {
        byte b = Receive2CharToByte();
        unsigned int RunningTime = (unsigned int)b * 10 ; //multiply to 10mS resolution
        //Make pivot turn
        TurnRight( RunningTime );
        Serial.println(";");
      }
      else if ( c == 'R' )
      {
        byte b = Receive2CharToByte();
        if ( b > 0 )
        {
          //if continuous running state is running forward
          if ( StateContRun == CONT_FORWARD )
          {
            //steer right (put steering factor value*2 to the left motor)
            byte Steer = b * 2 ;
            MyMBmotor.ProceedForward( Steer, 0 ) ;
          }
          else
          {
            //Make pivot turn
            unsigned int RunningTime = GetTurnTimeFromDeg( (unsigned int)b ) ;
            TurnRight( RunningTime );
          }
        }
        else
        {
          //if still running then wait for stop since it may turning oposit side
          MyMBmotor.WaitForStop() ;
          StateContRun = CONT_RIGHT ;
          //When continuous running starts, the choreography must be stopped
          FlagChoreograph = false ;
          MyMBmotor.RotateRight() ;
        }
        Serial.println(";");
      }
      else if ( c == 'l' )
      {
        byte b = Receive2CharToByte();
        unsigned int RunningTime = (unsigned int)b * 10 ; //multiply to 10mS resolution
        //Make pivot turn
        TurnLeft( RunningTime );
        Serial.println(";");
      }
      else if ( c == 'L' )
      {
        byte b = Receive2CharToByte();
        if ( b > 0 )
        {
          //if continuous running state is running forward
          if ( StateContRun == CONT_FORWARD )
          {
            //steer right (put steering factor value*2 to the left motor)
            byte Steer = b * 2 ;
            MyMBmotor.ProceedForward( 0, Steer ) ;
          }
          else
          {
            //Make pivot turn
            unsigned int RunningTime = GetTurnTimeFromDeg( (unsigned int)b ) ;
            TurnLeft( RunningTime );
          }
        }
        else
        {
          //if still running then wait for stop since it may turning oposit side
          MyMBmotor.WaitForStop() ;
          StateContRun = CONT_LEFT ;
          //When continuous running starts, the choreography must be stopped
          FlagChoreograph = false ;
          MyMBmotor.RotateLeft() ;
        }
        Serial.println(";");
      }
      else if ( c == 'E' )
      {
        int Adr = (int)Receive2CharToByte() * 16 ;
        Serial.println(";");
        for (byte a=0; a<16; a++)
        {
          byte b = EEPROM.read(Adr) ;
          char string[4] ;
          sprintf ( string, "%02X,", b ) ;
          Serial.print(string);
          Adr++ ;
        }
        Serial.println();
      }
      else if ( c == 'A' )
      {
        FlagChoreograph = !FlagChoreograph ;
        //When the choreography starts, continuous running must be stopped
        if ( FlagChoreograph == true )
        {
          StateContRun = CONT_OFF ;
          MyMBmotor.SlowStop() ;
        }
        Serial.println(";");
        ChoreSeqCount = 0 ;
        //restore obstacle detection range, since it may changed
        ObstacleRange = (unsigned int)EEPROM.read(EEADR_OBSTACLE_RANGE) * 10 ;
      }
      else if ( c == 'Q' )
      {
        //read I2C address
        I2Cadr = Receive2CharToByte();
        char string[10] ;
        sprintf ( string, "%02X:", I2Cadr ) ;
        Serial.print(string) ;
        unsigned int reading = ReadSensor ( I2Cadr ) ;
        sprintf ( string, "%c%04d %02d", Id, reading, SensLevel ) ;
        Serial.println(string);
        Serial.println(";");
      }
      else if ( c == 'S' )
      {
        FlagSens = !FlagSens;
        Serial.println(";");
      }
      else if ( c == 'I' )
      {
        //read I2C address from serial
        I2Cadr = Receive2CharToByte();
        //read command from serial
        char com = ReceiveSerial();
    
        byte param1 ;
        byte param2 ;
        byte paramflag = 0 ;
        while ( 1 )
        {
          c = ReceiveSerial();
          //Terminate by LF code
          if ( c == '\n' ) { break ;}
          //Cancel CR code
          if ( c != '\r' )
          {
            if ( paramflag < 2 )
            {
              paramflag++ ;
              param1 <<= 4 ;
              param1 += Hascii2bin(c) ;
            }
            else
            {
              paramflag++ ;
              param2 <<= 4 ;
              param2 += Hascii2bin(c) ;
            }
          }
        }
        Serial.print(":");
        Wire.beginTransmission(I2Cadr);
        //send command
        Wire.write(com);
        //if there is some parameter
        if ( paramflag > 0 ) { Wire.write(param1); }
        if ( paramflag > 2 ) { Wire.write(param2); }
        Wire.endTransmission();
  
        char string[8] ;
        sprintf ( string, "%02X%c", I2Cadr, com ) ;
        Serial.print(string);
  
        if ( paramflag > 0 )
        {
          sprintf ( string, "%02X", param1 ) ;
          Serial.print(string);
        }
        if ( paramflag > 2 )
        {
          sprintf ( string, "%02X", param2 ) ;
          Serial.print(string);
        }
        Serial.println(";");
      }
      else if ( c == 'V' )
      {
        Serial.println(";");
        Serial.println(VERSION);
      }
      else if ( c == 'X' )
      {
        Serial.print(";");
        //read battery voltage
        unsigned int BattVolt = MyMBmotor.ReadBatt() ;
        float Voltage = (float)BattVolt * 5 / 1024 ;
        Serial.print(Voltage);
        Serial.println("V");
      }
      else if ( c == 'Y' )
      {
        Serial.print(";");
        //show lowest battery voltage
        unsigned int BattVolt = MyMBmotor.ReadLowBatt() ;
        float Voltage = (float)BattVolt * 5 / 1024 ;
        Serial.print(Voltage);
        Serial.println("V");
        //reset voltage
        //MyMBmotor.InitLowBatt() ;
      }
      else if ( c == 'T' )
      {
        MyMBmotor.SlowStop() ;
        StateContRun = CONT_OFF ;
        //Stop the choreograph too
        FlagChoreograph = false ;
        Serial.println(";");
      }
      else
      {
        Serial.println("?");
      }
    }
    //the letter other than '$' is construed as RC command
    else
    {
      //renew the time of the command was received
      TimeRCcommand = millis() ;
      //Process RC command when it is changed
      if ( c != PrevRCcommand )
      {
        //Stop choreograph at any command
        //FlagChoreograph = false ;

        if ( c == RC_COM_FORWARD || c == RC_COM_STEERL || c == RC_COM_STEERR )
        {
          //start continuous run
          StateContRun = CONT_FORWARD ;
        }
        else
        {
          if ( PrevRCcommand == RC_COM_TURNR || PrevRCcommand == RC_COM_TURNL )
          {
            //Stop instantly if it is after the pivot turn
            MyMBmotor.InstantStop() ;
          }
          else
          {
            //Otherwise stop slowly
            MyMBmotor.WaitForStop() ;
          }
          StateContRun = CONT_OFF ;
        }

        PrevRCcommand = c ;

        if ( c == RC_COM_FORWARD )
        {
          MyMBmotor.ProceedForward(0,0) ;
        }
        else if ( c == RC_COM_STEERL )
        {
          MyMBmotor.ProceedForward( 0, RC_STEER_FACTOR ) ;
        }
        else if ( c == RC_COM_STEERR )
        {
          MyMBmotor.ProceedForward( RC_STEER_FACTOR, 0 ) ;
        }
        else if ( c == RC_COM_TURNL )
        {
          //Make pivot turn
          MyMBmotor.RotateLeft() ;
          StateContRun = CONT_LEFT ;
        }
        else if ( c == RC_COM_TURNR )
        {
          //Make pivot turn
          MyMBmotor.RotateRight() ;
          StateContRun = CONT_RIGHT ;
        }
        else if ( c == RC_COM_ACTION )
        {
          FlagChoreograph = !FlagChoreograph ;
        }
      }
    }
  }
  //See time out of the RC command
  if ( TimeRCcommand != 0 && millis() > TimeRCcommand + RC_TIMEOUT )
  {
    if ( PrevRCcommand == RC_COM_TURNR || PrevRCcommand == RC_COM_TURNL )
    {
      //Stop instantly if it is after the pivot turn
      MyMBmotor.InstantStop() ;
    }
    else
    {
      //Otherwise stop slowly
      MyMBmotor.WaitForStop() ;
    }
    StateContRun = CONT_OFF ;
    //Clear the previous-command buffer
    PrevRCcommand = ' ' ;
    //clear the time of RC command
    TimeRCcommand = 0 ;
  }

  //if it is in continuous running or doing choregraph then check battery
  if ( StateContRun != CONT_OFF || FlagChoreograph == true )
  {
    if ( CheckBattery() == false )
    {
      MyMBmotor.WaitForStop() ;
      StateContRun = CONT_OFF ;
      FlagChoreograph = false ;
      Serial.println(" Low battery");
    }
  }
  //if continuous forward running state, then check obstacle
  if ( StateContRun == CONT_FORWARD )
  {
    unsigned int Reading = ReadSensor ( I2CADR0 ) ;
    //if there is obstacle
    if ( Reading <= ObstacleRange )
    {
      MyMBmotor.SlowStop() ;
      StateContRun = CONT_OFF ;
      Serial.println(" Obstacle found");
    }
  }

  /***********************************************
   *Choreograph Sequence Proccess
  */
  //If choreograph flag is on
  if ( FlagChoreograph == true )
  {
    byte ChoreCommand = EEPROM.read(ChoreSeqAdr) ;
    Serial.print(ChoreSeqAdr, HEX);
    Serial.print(":");
    ChoreSeqAdr++ ;
    //if the command is no operation
    if ( ChoreCommand == 0xFF )
    {
      //do nothing
      Serial.println("Nop");
    }
    //if the command is restart then reset sequence address
    else if ( ChoreCommand == 0xF0 )
    {
      ChoreSeqAdr = CHORESEQADR ;
      ChoreSeqCount++ ;
      Serial.print("Choreograph count:");
      char string[8] ;
      sprintf ( string, "%d", ChoreSeqCount ) ;
      Serial.println(string);
      Serial.println("Restart");
    }
    //if the command is terminator then cease choreograph
    else if ( ChoreCommand == 0xF1 )
    {
      FlagChoreograph = false ;
      Serial.println("End");
    }
    else if ( ChoreCommand == 0xFE )
    {
      Serial.print("Scanning obstacle:");
      DistanceBuf = ReadSensor ( I2CADR0 ) ;
      char string[8] ;
      sprintf ( string, "%umm", DistanceBuf ) ;
      Serial.println(string);
    }
    else
    {
      //other commands has 1 parameter
      byte ChoreParam = EEPROM.read(ChoreSeqAdr) ;
      ChoreSeqAdr++ ;
      if ( ChoreCommand == 0x00 )
      {
        ObstacleRange = (unsigned int)ChoreParam * 10 ; //10mm resolution
        Serial.print("Detection range changed:");
        char string[8] ;
        sprintf ( string, "%dmm", ObstacleRange ) ;
        Serial.println(string);
      }
      else if ( ChoreCommand == 0x01 )
      {
        ChoreStepTime = ChoreParam ;
        Serial.print("Stepping time changed:");
        Serial.print(ChoreStepTime, DEC);
        Serial.println("00mS");
      }
      else if ( ChoreCommand == 0x10 )
      {
        ChoreSeqAdr = (unsigned int)ChoreParam ;
        Serial.print("Jump:");
        Serial.println(ChoreSeqAdr, HEX);
      }
      else if ( ChoreCommand == 0x11 )
      {
        ChoreSeqAdr = (unsigned int)ChoreParam + 256 ;
        Serial.print("Jump:");
        Serial.println(ChoreSeqAdr, HEX);
      }
      else if ( ChoreCommand == 0x12 )
      {
        ChoreSeqAdr = (unsigned int)ChoreParam + 512 ;
        Serial.print("Jump:");
        Serial.println(ChoreSeqAdr, HEX);
      }
      else if ( ChoreCommand == 0x13 )
      {
        ChoreSeqAdr = (unsigned int)ChoreParam  + 768 ;
        Serial.print("Jump:");
        Serial.println(ChoreSeqAdr, HEX);
      }
      else if ( ChoreCommand == 0x14 )
      {
        if ( DistanceBuf < ObstacleRange )
        {
          ChoreSeqAdr = (unsigned int)ChoreParam ;
          Serial.print("Obstacle found:Jump:");
          Serial.println(ChoreSeqAdr, HEX);
        }
        else
        {
          Serial.println("No obstacle");
        }
      }
      else if ( ChoreCommand == 0x15 )
      {
        if ( DistanceBuf < ObstacleRange )
        {
          ChoreSeqAdr = (unsigned int)ChoreParam + 256 ;
          Serial.print("Obstacle found:Jump:");
          Serial.println(ChoreSeqAdr, HEX);
        }
        else
        {
          Serial.println("No obstacle");
        }
      }
      else if ( ChoreCommand == 0x16 )
      {
        if ( DistanceBuf < ObstacleRange )
        {
          ChoreSeqAdr = (unsigned int)ChoreParam + 512 ;
          Serial.print("Obstacle found:Jump:");
          Serial.println(ChoreSeqAdr, HEX);
        }
      }
      else if ( ChoreCommand == 0x17 )
      {
        if ( DistanceBuf < ObstacleRange )
        {
          ChoreSeqAdr = (unsigned int)ChoreParam + 768 ;
          Serial.print("Obstacle found:Jump:");
          Serial.println(ChoreSeqAdr, HEX);
        }
        else
        {
          Serial.println("No obstacle");
        }
      }
      else if ( ChoreCommand == 0x20 )
      {
        unsigned int PauseTime = (unsigned int)ChoreParam * 100 ;
        Serial.print("Pause:");
        Serial.println(PauseTime, HEX);
        delay(PauseTime) ;
      }
      else if ( ChoreCommand == 0x21 )
      {
        unsigned int RunningTime = GetProcTimeFromCm( (unsigned int)ChoreParam ) ;
        Serial.print("Proceeding forward:");
        Serial.println(ChoreParam, HEX);
        GoForward( RunningTime );
      }
      else if ( ChoreCommand == 0x22 )
      {
        unsigned int TurnTimeMilli = GetTurnTimeFromDeg( (unsigned int)ChoreParam ) ;
        Serial.print("Turning right:");
        Serial.println(ChoreParam, HEX);
        TurnRight( TurnTimeMilli ) ;
      }
      else if ( ChoreCommand == 0x23 )
      {
        unsigned int TurnTimeMilli = GetTurnTimeFromDeg( (unsigned int)ChoreParam ) ;
        Serial.print("Turning left:");
        Serial.println(ChoreParam, HEX);
        TurnLeft( TurnTimeMilli ) ;
      }
    }
    //wait for the assigned stepping time past
    unsigned int StepTime = (unsigned int)ChoreStepTime * 100 ;
    delay (StepTime) ;
  }
  /***********************************************
   * Read range sensor value for the test purposes
  */
  if ( FlagSens == true )
  {
    if ( millis() > lastTime + 1000 )
    {
      lastTime = millis();
      char string[10] ;
      sprintf ( string, "%02X:", I2Cadr ) ;
      Serial.print(string) ;
      unsigned int reading = ReadSensor ( I2Cadr ) ;
      sprintf ( string, "%c%04d %02d", Id, reading, SensLevel ) ;
      Serial.println(string);
    }
  }
}