/*
  Motor control library for US droid mini robot

  Created by Steve Nishiarai 18/Jan/2019
  Copyright (c) 2019 Parawind Lab.  All right reserved.
*/

// include core Wiring API
#include "Arduino.h"

// include this library's description file
#include "MBmotor.h"

//reference offset. This is for the purpose of avoiding the calculated result become minus when comparing the BEMF voltage
#define REFOFFSET 256
//BEMF level of Stop
#define SPEED0 0

/*Motor BEMF feedback process constants
 * Motor is off for this time from the start of driving period to BEMF sensing AD conversion.
 * By the result of sensing, the motor PWM driving duty adjusted.

    |<---- motor BEMF sensing period ---->|
    |<-- sensing delay -->|<- PWM drive ->|
    |                     |               |
on
off |_____________________|_|_|_|_|_|_|_|_|_ <- motor drive PWM pulses
*/
//BEMF sensing period in milli second
#define SENSPERIOD 6
//BEMF sensing delay time in milli second
#define SENSDELAY 1

//parameter registers
// Motor driving trim when running forward
byte MR_Ftrim ;
byte ML_Ftrim ;
// Motor driving trim when turning
byte MR_Ttrim ;
byte ML_Ttrim ;
// Minimum speed
byte MinSpeed ;
// Nominal speed
byte NomSpeed ;
// Acceleration time
byte AccTime ;
// Motor BEMF feedback gain
byte Gain ;

int MCLpin;
int MCRpin;
int MSLpin;
int MSRpin;
int MREFpin;

// AD conversion value is 0 - 1023 for 0V - 5V, 1 bit resolution is 0.0049V
//Trget BEMF level
volatile unsigned int MR_TgBEMF ;
volatile unsigned int ML_TgBEMF ;

//Motor running flag
// true: motor acceleration process take place in interrupt process
// false: motor deceleration process take place in interrupt process
boolean FlagRun = false ;
//register of the lowest battery voltage that ever since it was initialized
volatile unsigned int LBattVolt ;
//BEMF reference voltage register
volatile unsigned int BEMFref ;
//motor drive timing counter used in timer2 interrupt
volatile byte DriveTimingCount = 0 ;

//PWM puls width register for right motor (use b14 - b7 for the actual PWM pulse width value)
int MR_Power = 0 ;
//PWM puls width register for leftt motor
int ML_Power = 0 ;
/*
 * Acceleration is realized by increasing target BEMF value from minimum speed BEMF value to nominal speed BEMF value.
 * Periodic time of increase is calculated by following equation.
 *
 * n= diferrence value between BEMF value in minimum speed and nominal speed
 * T= AccTime * 10 (acceleration time in milli second)
 * pT (periodic time) = T/n
 */ 
//Acceleration BEMF diferrence buffer
unsigned int RAccBemfDifBuf ;
unsigned int LAccBemfDifBuf ;
//Acceleration BEMF diferrence counter
// the counter which increasing target BEMF in acceleration period by the interrupt process
volatile unsigned int RAccBemfDifCounter ;
volatile unsigned int LAccBemfDifCounter ;
//Acceleration period buffer
unsigned int RAccPeriBuf ;
unsigned int LAccPeriBuf ;
//Acceleration period counter
unsigned int RAccPeriCounter ;
unsigned int LAccPeriCounter ;

//Constructor
MBmotor::MBmotor(int MCLpinI, int MCRpinI, int MSLpinI, int MSRpinI, int MREFpinI)
{
  MCLpin = MCLpinI;
  MCRpin = MCRpinI;
  MSLpin = MSLpinI;
  MSRpin = MSRpinI;
  MREFpin = MREFpinI;
  //assign output pin
  pinMode(MCLpin, OUTPUT);
  pinMode(MCRpin, OUTPUT);
}

void MBmotor::begin(byte MR_FtrimI, byte ML_FtrimI, byte MR_TtrimI, byte ML_TtrimI, byte MinSpeedI, byte NomSpeedI, byte AccTimeI, byte GainI)
{
  MR_Ftrim = MR_FtrimI ;
  ML_Ftrim = ML_FtrimI ;
  MR_Ttrim = MR_TtrimI ;
  ML_Ttrim = ML_TtrimI ;
  MinSpeed = MinSpeedI ;
  NomSpeed = NomSpeedI ;
  AccTime = AccTimeI ;
  Gain = GainI ;
  //initialize lowest battery voltage register
  LBattVolt = ReadBatt() ;
  //reset motor drive pins
  analogWrite(MCLpin,0);
  analogWrite(MCRpin,0);

  MR_TgBEMF = SPEED0 ;
  ML_TgBEMF = SPEED0 ;

  //setup timer interrupt
  cli();//stop interrupts
  //set timer2 interrupt at 1kHz
  TCCR2A = 0;// set entire TCCR2A register to 0
  TCCR2B = 0;// same for TCCR2B
  TCNT2  = 0;//initialize counter value to 0
  // set compare match register for 8khz increments
  OCR2A = 249;// = (16*10^6) / (1000*64) - 1 (must be <256)
  // turn on CTC mode
  TCCR2A |= (1 << WGM21);
  // Set CS21 bit for 64 prescaler
  TCCR2B |= (1 << CS22);   
  // enable timer2 compare interrupt
  TIMSK2 |= (1 << OCIE2A);
  sei();//allow interrupts
}

void MBmotor::ProceedForward( byte Lsteer, byte Rsteer )
{
  unsigned int MinBEMF = MinSpeed + REFOFFSET ;
  //calculate nominal speed
  unsigned int TmpNomBEMF  = MinBEMF + (unsigned int)NomSpeed ;
  //add steering and trim factor to the nominal speed
  // and calculate actual BEMF differential value between nominal speed and minimum speed
  unsigned int AccelTime = (unsigned int)AccTime * 10 ;
  //put trim to the nominal speed BEMF value
  unsigned int NomBEMF = Trim(Rsteer, MR_Ftrim, TmpNomBEMF) ;
  //calcurate BEMF diferrence
  RAccBemfDifBuf = NomBEMF - MinBEMF ;
  //calcurate period counter
  RAccPeriBuf = AccelTime / RAccBemfDifBuf ;
  //force period counter to the minimum value of 1 for prevent underflow in the interrupt process
  if ( RAccPeriBuf == 0 ) { RAccPeriBuf = 1 ;}
  NomBEMF = Trim(Lsteer, ML_Ftrim, TmpNomBEMF) ;
  LAccBemfDifBuf = NomBEMF - MinBEMF ;
  LAccPeriBuf = AccelTime / LAccBemfDifBuf ;
  //force period counter to the minimum value of 1 for prevent underflow in the interrupt process
  if ( LAccPeriBuf == 0 ) { LAccPeriBuf = 1 ;}
  //Set run flag
  FlagRun = true ;
  //Set BEMF diferrence for the interrupt process
  RAccBemfDifCounter = RAccBemfDifBuf ;
  LAccBemfDifCounter = LAccBemfDifBuf ;
  //initialize counter for the interrupt process
  RAccPeriCounter = RAccPeriBuf ;
  LAccPeriCounter = LAccPeriBuf ;
  //start motor
  MR_TgBEMF = MinBEMF ;
  ML_TgBEMF = MinBEMF ;
}

void MBmotor::RotateRight()
{
  unsigned int MinBEMF = MinSpeed + REFOFFSET ;
  //calculate nominal speed
  unsigned int NomBEMF  = MinBEMF + (unsigned int)NomSpeed ;
  //add steering and trim factor to the nominal speed
  // and calculate actual BEMF differential value between nominal speed and minimum speed
  unsigned int AccelTime = (unsigned int)AccTime * 10 ;
  //calcurate acceleration power counter
  NomBEMF = Trim(0, ML_Ttrim, NomBEMF) ;
  LAccBemfDifBuf = NomBEMF - MinBEMF ;
  RAccBemfDifBuf = 0 ; //this is required for disable deceleration in the interrupt process
  LAccPeriBuf = AccelTime / LAccBemfDifBuf ;
  //force period counter to the minimum value of 1 for prevent underflow in the interrupt process
  if ( LAccPeriBuf == 0 ) { LAccPeriBuf = 1 ;}
  //Set run flag
  FlagRun = true ;
  //Set BEMF diferrence for the interrupt process
  LAccBemfDifCounter = LAccBemfDifBuf ;
  //initialize counter for the interrupt process
  LAccPeriCounter = LAccPeriBuf ;
  //start motor
  ML_TgBEMF = MinBEMF ;
}

void MBmotor::RotateLeft()
{
  unsigned int MinBEMF = MinSpeed + REFOFFSET ;
  //calculate nominal speed
  unsigned int NomBEMF  = MinBEMF + (unsigned int)NomSpeed ;
  //add steering and trim factor to the nominal speed
  // and calculate actual BEMF differential value between nominal speed and minimum speed
  unsigned int AccelTime = (unsigned int)AccTime  * 10 ;
  //calcurate acceleration power counter
  NomBEMF = Trim(0, MR_Ttrim, NomBEMF) ;
  RAccBemfDifBuf = NomBEMF - MinBEMF ;
  LAccBemfDifBuf = 0 ; //this is required for disable deceleration in the interrupt process
  RAccPeriBuf = AccelTime / RAccBemfDifBuf ;
  //force period counter to the minimum value of 1 for prevent underflow in the interrupt process
  if ( RAccPeriBuf == 0 ) { RAccPeriBuf = 1 ;}
  //Set run flag
  FlagRun = true ;
  //Set BEMF diferrence for the interrupt process
  RAccBemfDifCounter = RAccBemfDifBuf ;
  //initialize counter for the interrupt process
  RAccPeriCounter = RAccPeriBuf ;
  //start motor
  MR_TgBEMF = MinBEMF ;
}

/***********************************************
 * Set parameters to deceleration for slow stop
 *
 * ! this function does not wait for stop
 */
void MBmotor::SlowStop()
{
  //Reset run flag
  FlagRun = false ;
  //Set target BEMF for the interrupt process
  RAccBemfDifCounter = RAccBemfDifBuf ;
  LAccBemfDifCounter = LAccBemfDifBuf ;
  //initialize counter for the interrupt process
  RAccPeriCounter = RAccPeriBuf ;
  LAccPeriCounter = LAccPeriBuf ;
}

/***********************************************
 * Set parameters to deceleration for fast stop
 *
 * ! this function does not wait for stop
 */
void MBmotor::FastStop()
{
  //Reset run flag
  FlagRun = false ;
  //Set target BEMF for the interrupt process
  RAccBemfDifCounter = 0 ;
  LAccBemfDifCounter = 0 ;
}

/***********************************************
 * Instant stop of the motor
 *
 * ! this function does not wait for stop
 */
void MBmotor::InstantStop()
{
  //Reset run flag
  FlagRun = false ;
  //Set motor power directly
  MR_Power = 0 ;
  ML_Power = 0 ;
  MR_TgBEMF = SPEED0 ;
  ML_TgBEMF = SPEED0 ;
}

/***********************************************
 * Wait until both motors are stop
 *
 */
void MBmotor::WaitForStop()
{
  MBmotor::SlowStop() ;
  while(MR_TgBEMF > SPEED0 || ML_TgBEMF > SPEED0){}
}

/***********************************************
 * Check both motors are stop
 *
 */
boolean MBmotor::CheckStop()
{
  if ( MR_TgBEMF == SPEED0 && ML_TgBEMF == SPEED0 ) { return true ;}
  return false ;
}

/***********************************************
 * Reset initialize battery voltage
 *
 */
void MBmotor::InitLowBatt()
{
  LBattVolt = ReadBatt() ;
}

/***********************************************
 * Return lowest battery voltage that ever since it was initialized
 *
 */
unsigned int MBmotor::ReadLowBatt()
{
  return LBattVolt ;
}

/***********************************************
 * Return current battery voltage
 *
 */
unsigned int MBmotor::ReadBatt()
{
  //return battery voltage from ADC
  return analogRead(MREFpin) ;
}

/***********************************************
 *Trim motor BEMF target value
 */
unsigned int MBmotor::Trim( byte steer, byte trim, unsigned int Speed )
{
  float x = (float)trim ;
  if ( x >= 128 ) { x -= 256 ;}
  x += (float)steer ;
//  x = (float)Speed * (1 + x/1000) ;
  x = (float)Speed * (1 + x/4096) ;
  Speed = (unsigned int)x ;
  return Speed ;
}

/***********************************************
 * motor drive timer interrupt processing routine
 */
ISR(TIMER2_COMPA_vect)
{
  //Motor on/off and BEMF sensing process
  if ( DriveTimingCount == 0 )
  {
    //Off motor power for BEMF sensing
    analogWrite(MCLpin,0);
    analogWrite(MCRpin,0);
  }
  else if ( DriveTimingCount == SENSDELAY )
  {
    //read battery voltage for the BEMF reference
    unsigned int Voltage = analogRead(MREFpin) ;
    //renew lowest voltage
    if ( Voltage < LBattVolt ) { LBattVolt = Voltage ;}
    //put offset to the reference voltage
    BEMFref = Voltage + REFOFFSET ;
    //put BEMF value to the monitoring buffer
    unsigned int MR_BEMFValue = BEMFref - analogRead(MSRpin) ;
    //if BEMF voltage is lower than target, that means motor is slow
    if ( MR_BEMFValue < MR_TgBEMF )
    {
      MR_Power += (int)Gain ;
      //put limitter
      if ( MR_Power < 0 ) { MR_Power = 0x07FFF ;}
    }
    else
    {
      MR_Power -= (int)Gain ;
      //put limitter
      if ( MR_Power < 0 ) { MR_Power = 0 ;}
    }
    //put BEMF value to the monitoring buffer
    unsigned int ML_BEMFValue = BEMFref - analogRead(MSLpin) ;
    //if BEMF value is smaller than target, that means motor is slow
    if ( ML_BEMFValue < ML_TgBEMF )
    {
      ML_Power += (int)Gain ;
      //put limitter
      if ( ML_Power < 0 ) { ML_Power = 0x07FFF ;}
    }
    else
    {
      ML_Power -= (int)Gain ;
      //put limitter
      if ( ML_Power < 0 ) { ML_Power = 0 ;}
    }
    //put next power to the motors
    int TempPower = MR_Power >> 7 ;
    byte PWMvalue = (byte)TempPower ;
    analogWrite(MCRpin,PWMvalue);
    TempPower = ML_Power >> 7 ;
    PWMvalue = (byte)TempPower ;
    analogWrite(MCLpin,PWMvalue);
  }
  if ( ++DriveTimingCount == SENSPERIOD ) { DriveTimingCount = 0 ;}

  //Acceleration/deceleration process
  if ( RAccBemfDifCounter == 0 )
  {
    if ( FlagRun == false )
    {
      //if deceleration time counter is 0 then stop
      MR_TgBEMF = SPEED0 ;
    }
  }
  else
  {
    RAccPeriCounter-- ;
    if ( RAccPeriCounter == 0 )
    {
      RAccPeriCounter = RAccPeriBuf ;
      RAccBemfDifCounter-- ;
      if ( FlagRun == true )
      {
        //acceleration
        MR_TgBEMF++ ;
      }
      else
      {
        //if target BEMF is still not stop speed ( to prevent wrap around )
        if ( MR_TgBEMF != SPEED0 )
        {
          //deceleration (stop at last)
          if ( RAccBemfDifCounter != 0 ) { MR_TgBEMF-- ;} else { MR_TgBEMF = SPEED0 ;}
        }
      }
    }
  }
  if ( LAccBemfDifCounter == 0 )
  {
    if ( FlagRun == false )
    {
      //if deceleration time counter is 0 then stop for the safe reason
      ML_TgBEMF = SPEED0 ;
    }
  }
  else
  {
    LAccPeriCounter-- ;
    if ( LAccPeriCounter == 0 )
    {
      LAccPeriCounter = LAccPeriBuf ;
      LAccBemfDifCounter-- ;
      if ( FlagRun == true )
      {
        //acceleration
        ML_TgBEMF++ ;
      }
      else
      {
        //if target BEMF is still not stop speed ( to prevent wrap around )
        if ( ML_TgBEMF != SPEED0 )
        {
          //deceleration (stop at last)
          if ( LAccBemfDifCounter != 0 ) { ML_TgBEMF-- ;} else { ML_TgBEMF = SPEED0 ;}
        }
      }
    }
  }
}
